/*
 * libetc - Copyright (C) 2005-2006 Luc Dufresne <luc@ordiluc.net>
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * See the file COPYING
 *
 */

#define _GNU_SOURCE
#include <dlfcn.h>
#include <stdio.h>
#include <sys/types.h>
#include <string.h>
#include <dirent.h>
#include <errno.h>
#include <sys/socket.h>
#include <sys/un.h>

#include "libetc.h"

#define ETCDIR ".config"
//#define DEBUG 1

FILE * (*orig_fopen) (const char *path, const char *mode);
FILE * (*orig_fopen64) (const char *path, const char *mode);
FILE * (*orig_freopen) (const char *path, const char *mode, FILE *stream);
FILE * (*orig_freopen64) (const char *path, const char *mode, FILE *stream);
int (*orig_open) (const char *pathname, int flags, mode_t mode);
int (*orig_open64) (const char *pathname, int flags, mode_t mode);
int (*orig_mkdir) (const char *pathname, mode_t mode);
int (*orig_rename) (const char *oldpath, const char *newpath);
int (*orig_unlink) (const char *pathname);
int (*orig_link) (const char *oldpath, const char *newpath);
int (*orig_symlink) (const char *oldpath, const char *newpath);
DIR * (*orig_opendir) (const char *name);
int (*orig_chdir) (const char *path);
int (*orig___xstat) (int ver, const char *file_name, struct stat *buf);
int (*orig___xstat64) (int ver, const char *file_name, struct stat64 *buf);
int (*orig___lxstat) (int ver, const char *file_name, struct stat *buf);
int (*orig___lxstat64) (int ver, const char *file_name, struct stat64 *buf);
int (*orig_access) (const char *pathname, int mode);
int (*orig_chmod) (const char *path, mode_t mode);
int (*orig_chown) (const char *path, uid_t owner, gid_t group);
int (*orig_lchown) (const char *path, uid_t owner, gid_t group);
int (*orig_utime) (const char *filename, struct utimbuf *buf);
int (*orig_utimes) (const char *filename, struct timeval *tvp);
int (*orig_truncate) (const char *path, off_t length);
int (*orig_truncate64) (const char *path, off_t length);
int (*orig___xmknod) (int ver, const char *pathname, mode_t mode, dev_t dev);
int (*orig_mkfifo) (const char *pathname, mode_t mode);
int (*orig_creat) (const char *pathname, mode_t mode);
int (*orig_creat64) (const char *pathname, mode_t mode);
//int (*orig_connect) (int sockfd, const struct sockaddr *serv_addr, socklen_t addrlen);
//int (*orig_bind) (int sockfd, const struct sockaddr *my_addr, socklen_t addrlen);
int (*orig_readlink) (const char *path, char *buf, size_t bufsiz);
int (*orig_rmdir) (const char *pathname);
int (*orig_mkstemp) (char *template);
int (*orig_mkstemp64) (char *template);

char *orig, *home, *etcdir;

void start_up () {
	struct stat info;
	char *etc, *xdg_config_home;

	orig_fopen = dlsym (RTLD_NEXT, "fopen");	
	orig_fopen64 = dlsym (RTLD_NEXT, "fopen64");
	orig_freopen = dlsym (RTLD_NEXT, "freopen");
	orig_freopen64 = dlsym (RTLD_NEXT, "freopen64");
	orig_open = dlsym (RTLD_NEXT, "open");
	orig_open64 = dlsym (RTLD_NEXT, "open64");
	orig_mkdir = dlsym (RTLD_NEXT, "mkdir");
	orig_rename = dlsym (RTLD_NEXT, "rename");
	orig_unlink = dlsym (RTLD_NEXT, "unlink");
	orig_link = dlsym (RTLD_NEXT, "link");
	orig_symlink = dlsym (RTLD_NEXT, "symlink");
	orig_opendir = dlsym (RTLD_NEXT, "opendir");
	orig_chdir = dlsym (RTLD_NEXT, "chdir");
	orig___xstat = dlsym (RTLD_NEXT, "__xstat");
	orig___xstat64 = dlsym (RTLD_NEXT, "__xstat64");
	orig___lxstat = dlsym (RTLD_NEXT, "__lxstat");
	orig___lxstat64 = dlsym (RTLD_NEXT, "__lxstat64");
	orig_access = dlsym (RTLD_NEXT, "access");
	orig_chmod = dlsym (RTLD_NEXT, "chmod");
	orig_chown = dlsym (RTLD_NEXT, "chown");
	orig_lchown = dlsym (RTLD_NEXT, "lchown");
	orig_utime = dlsym (RTLD_NEXT, "utime");
	orig_utimes = dlsym (RTLD_NEXT, "utimes");
	orig_truncate = dlsym (RTLD_NEXT, "truncate");
	orig_truncate64 = dlsym (RTLD_NEXT, "truncate64");
	orig___xmknod = dlsym (RTLD_NEXT, "__xmknod");
	orig_mkfifo = dlsym (RTLD_NEXT, "mkfifo");
	orig_creat = dlsym (RTLD_NEXT, "creat");
	orig_creat64 = dlsym (RTLD_NEXT, "creat64");
//	orig_connect = dlsym (RTLD_NEXT, "connect");
//	orig_bind = dlsym (RTLD_NEXT, "bind");
	orig_readlink = dlsym (RTLD_NEXT, "readlink");
	orig_rmdir = dlsym (RTLD_NEXT, "rmdir");
	orig_mkstemp = dlsym (RTLD_NEXT, "mkstemp");
	orig_mkstemp64 = dlsym (RTLD_NEXT, "mkstemp64");

	home = getenv ("HOME");
	orig = malloc (strlen (home) + 3);
	sprintf (orig, "%s/.", home);

	if (!(xdg_config_home = getenv ("XDG_CONFIG_HOME"))) {
		if (!(etc = getenv ("ETC"))) {
			etc = strdup (ETCDIR);
#ifdef DEBUG
			fprintf (stderr, "default value: %s\n", etc);
#endif
		}
#ifdef DEBUG
		fprintf (stderr, "$ETC: %s\n", etc);
#endif
		etcdir = malloc (strlen (home) + strlen (etc) + 1);
		sprintf (etcdir, "%s/%s", home, etc);
	} else {
#ifdef DEBUG
		fprintf (stderr, "$XDG_CONFIG_HOME: %s\n", xdg_config_home);
#endif
		etcdir = xdg_config_home;
	}
	
	if (orig___xstat (_STAT_VER, etcdir, &info)) {
		if (orig_mkdir (etcdir, 0700)) {
			fprintf (stderr, "Unable to create config directory %s: ", etcdir);
			perror ("");
			exit (2);
		}
	} else {
		if (!S_ISDIR(info.st_mode)) {
			fprintf (stderr, "ERROR: %s exists and is not a directory\n", etcdir);
			exit (2);
		}
	}
#ifdef DEBUG
	fprintf (stderr, "etcdir: %s\n", etcdir);
#endif
}

char *get_current_dir_name(void);

char *translate (const char *filename) {
	char *newfilename, *tmpfilename, *wd;

	if (!orig) start_up();
	wd = get_current_dir_name ();

	if ((0 == strcmp (wd, home)) // if cwd == $HOME
	    && filename [0] == '.'   // and a dotfile
	    && (0 != strcmp (filename, "."))
	    && (0 != strncmp (filename, "./", 2))
	    && (0 != strncmp (filename, "..", 2))) {
		tmpfilename = malloc (strlen (home) + strlen (filename) + 2);
		sprintf (tmpfilename, "%s/%s", home, filename);
		if (0 == strncmp (tmpfilename, etcdir, strlen(etcdir))) { // do not translate if trying to read/write in $XDG_CONFIG_HOME
			newfilename = strdup (filename);
		} else {
			filename++; // remove the dot
			newfilename = malloc (strlen (filename) + strlen (etcdir) + 2);
			sprintf (newfilename, "%s/%s", etcdir, filename);
#ifdef DEBUG
			fprintf (stderr, "RENAMED IN $HOME --> %s\n", newfilename);
#endif
		}
		free (tmpfilename);
	} else if (0 == strncmp (filename, orig, strlen (orig)) // if file name is $HOME/.something
		   && 0!= strncmp (filename, etcdir, strlen (etcdir)) ) { // do not translate if trying to read/write in $XDG_CONFIG_HOME
		filename += strlen (home) + 2; // remove "$HOME/." from the filename
		newfilename = malloc (strlen (filename) + strlen (etcdir) + 2);
		sprintf (newfilename, "%s/%s", etcdir, filename);
#ifdef DEBUG
		fprintf (stderr, "RENAMED --> %s\n", newfilename);
#endif
	} else { // not a dotfile
		newfilename = strdup (filename);
	}
	
	free (wd);
	return newfilename;
}

FILE* fopen (const char *path, const char *mode) {
	FILE *fichier;
	char *new_path;

#ifdef DEBUG
	fprintf (stderr, "fopen: %s\n", path);
#endif
	new_path = translate (path);
	fichier = orig_fopen (new_path, mode);
	free (new_path);
	return fichier;
}

FILE* fopen64 (const char *path, const char *mode) {
	FILE *fichier;
	char *new_path;

#ifdef DEBUG
	fprintf (stderr, "fopen64: %s\n", path);
#endif
	new_path = translate (path);
	fichier = orig_fopen64 (new_path, mode);
	free (new_path);
	return fichier;
}
				                                                      
FILE *freopen(const char *path, const char *mode, FILE *stream) {
	FILE *fichier;
	char *new_path;

#ifdef DEBUG
	fprintf (stderr, "freopen: %s\n", path);
#endif
	new_path = translate (path);
	fichier = orig_freopen (new_path, mode, stream);
	free (new_path);
	return fichier;
}

FILE *freopen64(const char *path, const char *mode, FILE *stream) {
	FILE *fichier;
	char *new_path;

#ifdef DEBUG
	fprintf (stderr, "freopen64: %s\n", path);
#endif
	new_path = translate (path);
	fichier = orig_freopen64 (new_path, mode, stream);
	free (new_path);
	return fichier;
}

int open (const char *pathname, int flags, mode_t mode) {
	int fichier;
	char *new_pathname;

#ifdef DEBUG
	fprintf (stderr, "open: %s\n", pathname);
#endif
	new_pathname = translate (pathname);
	fichier = orig_open (new_pathname, flags, mode);
	free (new_pathname);
	return fichier;
}

int open64 (const char *pathname, int flags, mode_t mode) {
	int fichier;
	char *new_pathname;

#ifdef DEBUG
	fprintf (stderr, "open64: %s\n", pathname);
#endif
	new_pathname = translate (pathname);
	fichier = orig_open64 (new_pathname, flags, mode);
	free (new_pathname);
	return fichier;
}       

int mkdir (const char *pathname, mode_t mode) {
	int return_value;
	char *new_pathname;

#ifdef DEBUG
	fprintf (stderr, "mkdir: %s\n", pathname);
#endif
	new_pathname = translate (pathname);
	return_value = orig_mkdir (new_pathname, mode);
	free (new_pathname);
	return return_value;
}

int rename (const char *oldpath, const char *newpath) {
	int return_value;
	char *new_newpath, *new_oldpath;

#ifdef DEBUG
	fprintf (stderr, "rename: %s %s\n", oldpath, newpath);
#endif
	new_oldpath = translate (oldpath);
	new_newpath = translate (newpath);
	return_value = orig_rename (new_oldpath, new_newpath);
	free (new_oldpath);
	free (new_newpath);
	return return_value;
}

int unlink (const char *pathname) {
	int return_value;
	char *new_pathname;

#ifdef DEBUG
	fprintf (stderr, "unlink: %s\n", pathname);
#endif
	new_pathname = translate (pathname);
	return_value = orig_unlink (new_pathname);
	free (new_pathname);
	return return_value;
}

int link (const char *oldpath, const char *newpath) {
	int return_value;
	char *new_newpath, *new_oldpath;

#ifdef DEBUG
	fprintf (stderr, "link: %s %s\n", oldpath, newpath);
#endif
	new_newpath = translate (newpath);
	new_oldpath = translate (oldpath);
	return_value = orig_link (new_oldpath, new_newpath);
	free (new_newpath);
	free (new_oldpath);
	return return_value;
}

int symlink (const char *oldpath, const char *newpath) {
	int return_value;
	char *new_newpath, *new_oldpath;

#ifdef DEBUG
	fprintf (stderr, "symlink: %s %s\n", oldpath, newpath);
#endif
	new_newpath = translate (newpath);
	new_oldpath = translate (oldpath);
	return_value = orig_symlink (new_oldpath, new_newpath);
	free (new_newpath);
	free (new_oldpath);
	return return_value;
}

DIR *opendir(const char *name) {
	DIR * return_value;
	char *new_name;

#ifdef DEBUG
	fprintf (stderr, "opendir: %s\n", name);
#endif
	new_name = translate (name);
	return_value = orig_opendir (new_name);
	free (new_name);
	return return_value;
}

int chdir (const char *path) {
	int return_value;
	char *new_path;

#ifdef DEBUG
	fprintf (stderr, "chdir: %s\n", path);
#endif
	new_path = translate (path);
	return_value = orig_chdir (new_path);
	free (new_path);
	return return_value;
}

int __xstat (int ver, const char *file_name, struct stat *buf) {
	int return_value;
	char *new_filename;

#ifdef DEBUG
	fprintf (stderr, "__xstat: %s\n", file_name);
#endif
	new_filename = translate (file_name);
	return_value = orig___xstat (ver, new_filename, buf);
	free (new_filename);
	return return_value;
}

int __xstat64 (int ver, const char *file_name, struct stat64 *buf) {
	int return_value;
	char *new_filename;

#ifdef DEBUG
	fprintf (stderr, "__xstat64: %s\n", file_name);
#endif
	new_filename = translate (file_name);
	return_value = orig___xstat64 (ver, new_filename, buf);
	free (new_filename);
	return return_value;
}

int __lxstat (int ver, const char *file_name, struct stat *buf) {
	int return_value;
	char *new_filename;

#ifdef DEBUG
	fprintf (stderr, "__lxstat: %s\n", file_name);
#endif
	new_filename = translate (file_name);
	return_value = orig___lxstat (ver, new_filename, buf);
	free (new_filename);
	return return_value;
}

int __lxstat64 (int ver, const char *file_name, struct stat64 *buf) {
	int return_value;
	char *new_filename;

#ifdef DEBUG
	fprintf (stderr, "__lxstat64: %s\n", file_name);
#endif
	new_filename = translate (file_name);
	return_value = orig___lxstat64 (ver, new_filename, buf);
	free (new_filename);
	return return_value;
}

int access(const char *pathname, int mode) {
	int return_value;
	char *new_pathname;

#ifdef DEBUG
	fprintf (stderr, "access: %s\n", pathname);
#endif
	new_pathname = translate (pathname);
	return_value = orig_access (new_pathname, mode);
	free (new_pathname);
	return return_value;
}

int chmod(const char *path, mode_t mode) {
	int return_value;
	char *new_path;

#ifdef DEBUG
	fprintf (stderr, "chmod: %s\n", path);
#endif
	new_path = translate (path);
	return_value = orig_chmod (new_path, mode);
	free (new_path);
	return return_value;
}

int chown(const char *path, uid_t owner, gid_t group) {
	int return_value;
	char *new_path;

#ifdef DEBUG
	fprintf (stderr, "chown: %s\n", path);
#endif
	new_path = translate (path);
	return_value = orig_chown (new_path, owner, group);
	free (new_path);
	return return_value;
}

int lchown(const char *path, uid_t owner, gid_t group) {
	int return_value;
	char *new_path;

#ifdef DEBUG
	fprintf (stderr, "lchown: %s\n", path);
#endif
	new_path = translate (path);
	return_value = orig_lchown (new_path, owner, group);
	free (new_path);
	return return_value;
}

int utime(const char *filename, struct utimbuf *buf) {
	int return_value;
	char *new_filename;

#ifdef DEBUG
	fprintf (stderr, "utime: %s\n", filename);
#endif
	new_filename = translate (filename);
	return_value = orig_utime (new_filename, buf);
	free (new_filename);
	return return_value;
}

int utimes(char *filename, struct timeval *tvp) {
	int return_value;
	char *new_filename;

#ifdef DEBUG
	fprintf (stderr, "utimes: %s\n", filename);
#endif
	new_filename = translate (filename);
	return_value = orig_utimes (new_filename, tvp);
	free (new_filename);
	return return_value;
}

int truncate(const char *path, off_t length) {
	int return_value;
	char *new_path;

#ifdef DEBUG
	fprintf (stderr, "utimes: %s\n", path);
#endif
	new_path = translate (path);
	return_value = orig_truncate (new_path, length);
	free (new_path);
	return return_value;
}

int truncate64 (const char *path, off_t length) {
	int return_value;
	char *new_path;

#ifdef DEBUG
	fprintf (stderr, "utimes: %s\n", path);
#endif
	new_path = translate (path);
	return_value = orig_truncate64 (new_path, length);
	free (new_path);
	return return_value;
}       

int __xmknod(int ver, const char *path, mode_t mode, dev_t dev) {
	int return_value;
	char *new_path;
	
#ifdef DEBUG
	fprintf (stderr, "mknod: %s\n", path);
#endif
	new_path = translate (path);
	return_value = orig___xmknod (ver, new_path, mode, dev);
	free (new_path);
	return return_value;
}
       
int mkfifo(const char *path, mode_t mode) {
	int return_value;
	char *new_path;

#ifdef DEBUG
	fprintf (stderr, "mkfifo: %s\n", path);
#endif
	new_path = translate (path);
	return_value = orig_mkfifo (new_path, mode);
	free (new_path);
	return return_value;
}       

int creat(const char *path, mode_t mode) {
	int return_value;
	char *new_path;

#ifdef DEBUG
	fprintf (stderr, "creat: %s\n", path);
#endif
	new_path = translate (path);
	return_value = orig_creat (new_path, mode);
	free (new_path);
	return return_value;
}

int creat64(const char *path, mode_t mode) {
	int return_value;
	char *new_path;

#ifdef DEBUG
	fprintf (stderr, "creat64: %s\n", path);
#endif
	new_path = translate (path);
	return_value = orig_creat64 (new_path, mode);
	free (new_path);
	return return_value;
}


/*
 * connect() and bind ()
 * 
 * I don't think it is needed.
 *
 * kde does:
 * readlink ("$HOME/.kde/socket...", foo, ...);
 * bind AF_LOCAL, foo (with foo = $TMP/ksocket...)
 *
 * so we just need to wrap readlink
 * 
 *
 * Do not uncomment unless the buffer overflow is corrected
 * 
int connect(int sockfd, const struct sockaddr *serv_addr, socklen_t addrlen) {
        struct sockaddr_un newaddr;
        char *new_path;
	int return_value;

	if (serv_addr->sa_family == AF_LOCAL) {
#ifdef DEBUG
		fprintf (stderr, "connect: %s\n", serv_addr->sa_data);
#endif
		newaddr.sun_family = AF_LOCAL;
		new_path = translate (serv_addr->sa_data);
		printf ("%s\n", new_path);
//TODO	
		strcpy (newaddr.sun_path, new_path); // buffer overflow !
		return_value = orig_connect (sockfd, &newaddr, sizeof (newaddr));
		free (new_path);
	} else {
//TODO
		return_value = orig_connect (sockfd, serv_addr, addrlen); // maybe orig_connect is not defined
	}       
	return return_value;
}

int bind(int sockfd, const struct sockaddr *my_addr, socklen_t addrlen) {
	struct sockaddr_un newaddr;
	char *new_path;
	int return_value;
	
	if (my_addr->sa_family == AF_LOCAL) {
#ifdef DEBUG
		fprintf (stderr, "bind: %s\n", my_addr->sa_data);
#endif
		newaddr.sun_family = AF_LOCAL;
		new_path = translate (my_addr->sa_data);
		printf ("%s\n", new_path);
//TODO
		strcpy (newaddr.sun_path, new_path); // buffer overflow !
		return_value = orig_bind (sockfd, &newaddr, sizeof (newaddr));
		free (new_path);
	} else {
//TODO
		return_value = orig_bind (sockfd, my_addr, addrlen); //maybe orig_bind is not defined
	}
	return return_value;
}
*/

int readlink(const char *path, char *buf, size_t bufsiz) {
	int return_value;
	char *new_path;

#ifdef DEBUG
	fprintf (stderr, "readlink: %s\n", path);
#endif
	new_path = translate (path);
	return_value = orig_readlink (new_path, buf, bufsiz);
	free (new_path);
	return return_value;
}

int rmdir(const char *pathname) {
        int return_value;
	char *new_pathname;

#ifdef DEBUG
	fprintf (stderr, "rmdir: %s\n", pathname);
#endif
	new_pathname = translate (pathname);
	return_value = orig_rmdir (new_pathname);
	free (new_pathname);
	return return_value;
	
}

int mkstemp(char *template) {
	int return_value, i;
	char *new_template;
	size_t taille1, taille2;

#ifdef DEBUG
	fprintf (stderr, "mkstemp: %s\n", template);
#endif
	new_template = translate (template);
	return_value = orig_mkstemp (new_template);
	taille1 = strlen (template);
	taille2 = strlen (new_template);
	for (i = 0; i <= 6; i++) {
		template[taille1 - i] = new_template[taille2 - i];
	}
	return return_value;
}

int mkstemp64(char *template) {
	int return_value, i;
	char *new_template;
	size_t taille1, taille2;

#ifdef DEBUG
	fprintf (stderr, "mkstemp64: %s\n", template);
#endif
	new_template = translate (template);
	return_value = orig_mkstemp64 (new_template);
	taille1 = strlen (template);
	taille2 = strlen (new_template);
	for (i = 0; i <= 6; i++) {
		template[taille1 - i] = new_template[taille2 - i];
	}
	return return_value;
}
